/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef XMLFILEHEADERFILE
#define XMLFILEHEADERFILE

#include <QtXml/QDomDocument>
#include <QtXml/QDomElement>
#include <QtXml/QDomNode>
#include <QString>

class TXmlFileException
{
public:

enum TExceptionType
{
	FILE_OPEN_ERROR=0,
	FILE_DATA_ERROR
};
// CREATORS
	TXmlFileException(const QString &Error,const TExceptionType Type=FILE_DATA_ERROR);
	~TXmlFileException(void);

// ACCESSORS
	QString ErrorText(void) const {return d_ErrorText;}
	TExceptionType ErrorType(void) const {return d_ErrorType;}

private:
	QString							d_ErrorText;
	TExceptionType					d_ErrorType;
};

class TXmlFile
{
public:
// CREATORS
	TXmlFile(void);
	~TXmlFile(void);

// ACCESSORS
	QString 		FileName(void) const {return d_FileName;}
	QDomElement* 	RootNode(void) {return &d_RootNode;}

	// Reading
	void 			Open(const QString &FileIdentification,int *VersionNumber=0,const bool process_namespace = true);
	QString 		DocumentType(void) const {return d_DocumentType;}	// for '!DOCTYPE'
	QString 		ReadTextNode(const QDomNode *Node,const bool Optional=false) const;
	QString 		ReadTextNode(const QDomElement *Node,const bool Optional=false) const;
	QString 		ReadAttribute(const QDomElement *Node,const QString &Name,const bool Optional=false) const;
	QString 		ReadAttribute(const QDomNode *Node,const QString &Name,const bool Optional=false) const;

	QDomElement 	GetNode(const QDomElement *Parent,const QString &Name,const bool Optional=false) const;
	QDomElement 	GetNode(const QDomNode *Parent,const QString &Name,const bool Optional=false) const;
	QDomElement 	GetSiblingNode(const QDomElement *CurrentNode,const QString &Name,const bool Optional=false) const;
	QDomNode	 	GetFirstNode(const QDomElement *Parent,const bool Optional=false) const;
	QDomNode	 	GetFirstNode(const QDomNode *Parent,const bool Optional=false) const;
	QDomNode	 	GetSiblingNode(const QDomNode *CurrentNode,const bool Optional=false) const;

// MANIPULATORS
	void 			SetFileName(const QString &FileName) {d_FileName = FileName;}
	QDomElement 	CreateNode(QDomElement *ParentNode,const QString &Name);
	void			Clear(void);	// clear all data

	// Writing
	void 			Reset(const QString &FileIdentification,const int VersionNumber);
	void 			Reset(const QString &DocType,const QString &FileIdentification,const int VersionNumber);
	void 			WriteTextNode(QDomElement *Node,const QString &Data);
	void 			WriteTextNode(QDomElement *Node,const double &Value,const int Precision=6);
	void 			WriteTextNode(QDomElement *Node,const int Value);
	void 			WriteAttribute(QDomElement *Node, const QString &Name,const QString &Attribute);
	void 			WriteAttribute(QDomElement *Node, const QString &Name,const double &Attribute,const int Precision=6);
	void 			WriteAttribute(QDomElement *Node, const QString &Name,const int Attribute);

	void 			WriteXmlFile(const bool inline_xml = false);

private:
	QDomDocument						d_XmlData;
	QDomElement							d_RootNode;
	QString								d_FileName;
	QString								d_FileIdentification;
	QString								d_DocumentType;

// NOT IMPLEMENTED
	TXmlFile(const TXmlFile&);
	TXmlFile& operator=(const TXmlFile&);
};

#endif
