/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef FILEDATAHEADERFILE
#define FILEDATAHEADERFILE

#include <QString>
#include <QStringList>
#include <QHash>
#include <vector>

class TXmlFile;

class TFileDataNode
{
public:

// CREATORS
	TFileDataNode(TFileDataNode *Parent=0);
	~TFileDataNode(void);

// ACCESSORS
	bool Is_Directory(void) const {return d_is_directory;}
	int Children_Count(void) const;
	TFileDataNode* Child(const int offset) const;
	QString File_Name(void) const {return d_file_name;}

// MANIPULATORS
	void Set_Directory(const bool b);
	void Add_Child(TFileDataNode*);
	void Set_File_Name(const QString &file_name);

	void Clear(void);

private:

	bool								d_is_directory;
	std::vector<TFileDataNode*>			d_children;
	QString								d_file_name;

// NOT IMPLEMENTED
	TFileDataNode(const TFileDataNode&);
	TFileDataNode& operator=(const TFileDataNode&);

	friend class 						TFileData;
};

class TFileData
{
public:
	
// CREATORS
	TFileData(void);
	~TFileData(void);

// ACCESSORS
	bool Save_File_List(const QString &source_path,const QString &file_list_name) const;

	int Update_File_Count(void) const;
	QString Get_Update_File_Name(const QString &path,const int offset) const;

	int Source_File_Count(void) const;
	QString Get_Source_File_Name(const QString &path,const int offset) const;
	
	QString Get_Last_Error(void) const {return d_last_error;}

// MANIPULATORS
	bool Load_File_List(const QString &path,const QString &file_list_name);
	bool Create_File_List(const QString &path);
	bool Create_Update_File_List(const QString &path);
	void Clear(void) {d_root_node.Clear();}

private:

	TFileDataNode						d_root_node;
	QStringList							d_source_file_names;
	QStringList							d_update_file_names;
	
    QHash<QString, unsigned int> 		d_source_file_hash;
	mutable QString						d_last_error;

	void Update_Source_File_Names(void);
	void Create_File_List(TFileDataNode *parent,const QString &path);
	void Update_Source_File_Names(const QString &Prefix,QStringList *list,TFileDataNode *node);
	void Load_File_List_V1(TXmlFile *xmlfile);

// NOT IMPLEMENTED
	TFileData(const TFileData&);
	TFileData& operator=(const TFileData&);

};

#endif
