/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QVariant>
#include <QAction>
#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QSpacerItem>
#include <QPlainTextEdit>
#include <QToolButton>
#include <QVBoxLayout>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QDir>
#include <QProcess>
#include <QDateTime>
#include <QSettings>
#include <QCloseEvent>

#include "../../core/messagebox.h"

#include "foldersync.h"

TFolderSync::TFolderSync(
	const QWidget						*parent,
	Qt::WindowFlags						flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QVBoxLayout							*dialog_vlayout;
    QGridLayout							*folder_layout;
    QLabel								*source_label;
    QLabel								*destination_label;
	QLabel								*footer_label;
	QFrame								*separator_upper_line;
    QFrame								*separator_middle_line;
	QFrame								*separator_lower_line;
    QHBoxLayout							*button_hlayout;
    QSpacerItem							*button_hspacer;
	QDir								data_path;
	QFont								text_font("courier new");
	QFont								footer_font("Verdana");
	
	text_font.setFixedPitch(true);
	
	footer_font.setPointSize(8);
	footer_font.setStyle(QFont::StyleItalic);
	
	data_path = QDir::home();
	
	if(!data_path.exists(".FolderSync"))
	{
		data_path.mkdir(".FolderSync");
	}
	
	data_path.cd(".FolderSync");
	d_data_path = data_path.absolutePath();
	
	d_settings = new QSettings(d_data_path + "/Settings.ini",QSettings::IniFormat);
	
	this->resize(601, 378);
	
	dialog_vlayout = new QVBoxLayout(this);

	folder_layout = new QGridLayout();

	source_label = new QLabel(this);
	folder_layout->addWidget(source_label, 0, 0, 1, 1);
	
	d_source_edit = new QLineEdit(this);
	folder_layout->addWidget(d_source_edit, 0, 1, 1, 1);
	
	d_source_browse_button = new QToolButton(this);
	folder_layout->addWidget(d_source_browse_button, 0, 2, 1, 1);
	
	destination_label = new QLabel(this);
	folder_layout->addWidget(destination_label, 1, 0, 1, 1);
	
	d_destination_edit = new QLineEdit(this);
	folder_layout->addWidget(d_destination_edit, 1, 1, 1, 1);
	
	d_destination_browse_button = new QToolButton(this);
	folder_layout->addWidget(d_destination_browse_button, 1, 2, 1, 1);
		
	dialog_vlayout->addLayout(folder_layout);
	
	separator_upper_line = new QFrame(this);
	separator_upper_line->setFrameShape(QFrame::HLine);
	separator_upper_line->setFrameShadow(QFrame::Sunken);
	dialog_vlayout->addWidget(separator_upper_line);

	d_file_log = new QPlainTextEdit(this);
	d_file_log->setWordWrapMode(QTextOption::NoWrap);
	d_file_log->setFont(text_font);
	dialog_vlayout->addWidget(d_file_log);
	
	separator_middle_line = new QFrame(this);
	separator_middle_line->setFrameShape(QFrame::HLine);
	separator_middle_line->setFrameShadow(QFrame::Sunken);
	dialog_vlayout->addWidget(separator_middle_line);
	
	button_hlayout = new QHBoxLayout();
	
	d_source_report_button = new QPushButton(this);
	d_source_report_button->setAutoDefault(false);
	button_hlayout->addWidget(d_source_report_button);
	
	d_sync_button = new QPushButton(this);
	d_sync_button->setAutoDefault(false);
	button_hlayout->addWidget(d_sync_button);

	button_hspacer = new QSpacerItem(40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);
	
	d_close_button = new QPushButton(this);
	d_close_button->setAutoDefault(false);
	button_hlayout->addWidget(d_close_button);
		
	dialog_vlayout->addLayout(button_hlayout);
	
	separator_lower_line = new QFrame(this);
	separator_lower_line->setFrameShape(QFrame::HLine);
	separator_lower_line->setFrameShadow(QFrame::Sunken);
	dialog_vlayout->addWidget(separator_lower_line);

	footer_label = new QLabel(this);
	footer_label->setFont(footer_font);
	dialog_vlayout->addWidget(footer_label);

	
	d_msg_box = new TMessageBox(this);
	
	QWidget::setTabOrder(d_source_edit, d_source_browse_button);
	QWidget::setTabOrder(d_source_browse_button, d_destination_edit);
	QWidget::setTabOrder(d_destination_edit, d_destination_browse_button);
	QWidget::setTabOrder(d_destination_browse_button, d_file_log);
	QWidget::setTabOrder(d_file_log, d_source_report_button);
	QWidget::setTabOrder(d_source_report_button, d_sync_button);
	QWidget::setTabOrder(d_sync_button, d_close_button);
	
	this->setWindowTitle(QStringLiteral("Controlled Documents Folder Sync - Version 1.3"));
	
	d_source_edit->setText(d_settings->value(QStringLiteral("Source"),QString()).toString());
	d_destination_edit->setText(d_settings->value(QStringLiteral("Destination"),QString()).toString());
	d_program_touch = d_settings->value(QStringLiteral("Program_Touch"),QString()).toString();
	d_program_touch_datetime_format = d_settings->value(QStringLiteral("Program_Touch_DateTime"),QString()).toString();
	
	if(d_program_touch.length() == 0)
	{
#ifdef Q_OS_WIN32
		d_program_touch = QStringLiteral("touch.exe");	// relative to application
#else
		d_program_touch = QStringLiteral("/usr/bin/touch");
#endif
	}
	
	if(d_program_touch_datetime_format.length() == 0)
	{
#ifdef Q_OS_WIN32
		d_program_touch_datetime_format = QStringLiteral("MMddhhmmyyyy.ss");	// touch.exe for windows uses this format
#else
		d_program_touch_datetime_format = QStringLiteral("yyyyMMddhhmm.ss");
#endif
	}

	source_label->setText(QStringLiteral("Source:"));
	d_source_browse_button->setText(QStringLiteral("..."));
	destination_label->setText(QStringLiteral("Destination:"));
	d_destination_browse_button->setText(QStringLiteral("..."));
	d_source_report_button->setText(QStringLiteral("Source Report"));
	d_sync_button->setText(QStringLiteral("Sync"));
	d_close_button->setText(QStringLiteral("Close"));
	footer_label->setText(QStringLiteral("Folder Sync © 2017 Select Calibration Incorporated.  www.selectcalibration.ca"));

	connect(d_source_report_button,SIGNAL(clicked(void)),this,SLOT(Source_Report(void)));
	connect(d_sync_button,SIGNAL(clicked(void)),this,SLOT(Sync(void)));
	connect(d_source_browse_button,SIGNAL(clicked(void)),this,SLOT(Browse_Source(void)));
	connect(d_destination_browse_button,SIGNAL(clicked(void)),this,SLOT(Browse_Destination(void)));
	connect(d_close_button,SIGNAL(clicked(void)),this,SLOT(close(void)));
}

TFolderSync::~TFolderSync(void)
{
	delete d_settings;
}

void TFolderSync::Sync(void)
{
	this->Initialize_Log();
	
	d_sync_button->setEnabled(false);
	d_source_report_button->setEnabled(false);
	
	if(this->Create_Reference())
	{
		if(this->Update_Target_Folder())
		{
			this->Write_Log();
			
			d_sync_button->setEnabled(true);
			d_source_report_button->setEnabled(true);
			
			return;
		}
	}
	
	d_sync_button->setEnabled(true);
	d_source_report_button->setEnabled(true);

	d_file_log->insertPlainText(QString("\n"));
	this->Add_Log_Entry("ERR:  Update did not complete.");
}

void TFolderSync::Source_Report(void)
{
	this->Initialize_Log();
	
	d_sync_button->setEnabled(false);
	d_source_report_button->setEnabled(false);

	if(!this->Create_Reference())
	{
		d_file_log->insertPlainText(QString("\n"));
		this->Add_Log_Entry("ERR:  Source Report did not complete.");
	}
	
	d_sync_button->setEnabled(true);
	d_source_report_button->setEnabled(true);
}

void TFolderSync::Browse_Source(void)
{
	QString								folder_name;
	QString								parent_folder;
	int									index;
	
	parent_folder = d_source_edit->text();
	if(parent_folder.length())
	{
		parent_folder.replace('\\','/');
	
		index = parent_folder.lastIndexOf('/');
		
		if(index > 0)
		{
			parent_folder.truncate(index);
		}
	}
	
	folder_name = QFileDialog::getExistingDirectory(this,"Select Source Folder",parent_folder);
	
	if(folder_name.length())
	{
		d_source_edit->setText(QDir::toNativeSeparators(folder_name));
	}
}

void TFolderSync::Browse_Destination(void)
{
	QString								folder_name;
	QString								parent_folder;
	int									index;
	
	parent_folder = d_destination_edit->text();
	if(parent_folder.length())
	{
		parent_folder.replace('\\','/');
		
		index = parent_folder.lastIndexOf('/');
		
		if(index > 0)
		{
			parent_folder.truncate(index);
		}
	}
	
	folder_name = QFileDialog::getExistingDirectory(this,"Select Destination Folder",parent_folder);
	
	if(folder_name.length())
	{
		d_destination_edit->setText(QDir::toNativeSeparators(folder_name));
	}
}

void TFolderSync::closeEvent(
	QCloseEvent 						*event)
{
	d_settings->setValue("Source", d_source_edit->text());
	d_settings->setValue("Destination", d_destination_edit->text());
	d_settings->setValue("Program_Touch", d_program_touch);
	d_settings->setValue("Program_Touch_DateTime", d_program_touch_datetime_format);
	
	event->accept();
}

bool TFolderSync::Create_Reference(void)
{
	QFileInfo							file_info;
	QString								path;
	QString								folder_name;
	QString								text;
	QDateTime							date_time;
	TFileData							reference_data;
	int									count;
	int									cntr;
	
	folder_name = QDir::fromNativeSeparators(d_source_edit->text());
	
	if(folder_name.endsWith('/'))
	{
		folder_name.chop(1);
	}
	
	file_info.setFile(folder_name);
	
	if(!file_info.exists())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The source directory does not exist.");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}

	if(!file_info.isDir())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The source must be a directory.");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	path = file_info.absoluteFilePath();
	
	if(!reference_data.Create_File_List(path))
	{
		d_msg_box->setText("Read Error");
		d_msg_box->setInformativeText("Cannot read information from source directory.");
		d_msg_box->setDetailedText(reference_data.Get_Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	if(!reference_data.Save_File_List(path,d_data_path + "/reference_list.xml"))
	{
		d_msg_box->setText("Write Error");
		d_msg_box->setInformativeText("Cannot write source file list.");
		d_msg_box->setDetailedText(reference_data.Get_Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	this->Add_Log_Entry("Controlled Files");
	count = reference_data.Source_File_Count();
	
	for(cntr = 0;cntr < count;++cntr)
	{
		text = reference_data.Get_Source_File_Name(QString(),cntr);
		file_info.setFile(path + text);
		
		date_time = file_info.lastModified();
		
		this->Add_Log_Entry(date_time.toString("dd MMM yyyy hh:mm:ss AP"),text);
	}

	return true;
}

bool TFolderSync::Update_Target_Folder(void)
{
	QFileInfo							file_info;
	int									cntr;
	QString								text;
	QString								source_folder;
	QString								destination_folder;
	QString								source_file;
	QString								destination_file;
	QProcess							process;
	QStringList							arguments;
	int									file_count;
	TFileData							reference_data;
	QStringList::const_iterator			iter;
	QDir								dir;
	QDateTime							date_time;

	source_folder = QDir::fromNativeSeparators(d_source_edit->text());
	
	if(source_folder.endsWith('/'))
	{
		source_folder.chop(1);
	}
	
	file_info.setFile(source_folder);
	
	if(!file_info.exists())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The source directory does not exist.");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	if(!file_info.isDir())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The source must be a directory");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}

	destination_folder = QDir::fromNativeSeparators(d_destination_edit->text());
	
	if(destination_folder.endsWith('/'))
	{
		destination_folder.chop(1);
	}
	
	file_info.setFile(destination_folder);
	
	if(!file_info.exists())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The destination directory does not exist.");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	if(!file_info.isDir())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The destination must be a directory");
		d_msg_box->setDetailedText(QString());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	source_folder += "/";
	destination_folder += "/";
	
	if(!reference_data.Load_File_List(d_data_path,"reference_list.xml"))
	{
		d_msg_box->setText("Read Error");
		d_msg_box->setInformativeText("Cannot read source file list");
		d_msg_box->setDetailedText(reference_data.Get_Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}

	if(!reference_data.Create_Update_File_List(destination_folder))
	{
		d_msg_box->setText("Compare Error");
		d_msg_box->setInformativeText("Comparison of the source and destination file list could not be done");
		d_msg_box->setDetailedText(reference_data.Get_Last_Error());
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return false;
	}
	
	d_TransferFileList.clear();
	d_TransferDirList.clear();
	
	file_count = reference_data.Update_File_Count();
	
	for(cntr = 0;cntr < file_count;++cntr)
	{
		text = reference_data.Get_Update_File_Name(QString(),cntr);
	
		d_TransferFileList.push_back(text);
		
		text.resize(text.lastIndexOf('/'));
		
		if(text.length())
		{
			this->Add_Transfer_Path(text);
		}
	}
	
	for(iter = d_TransferDirList.begin();iter < d_TransferDirList.end();++iter)
	{
		dir.setPath(destination_folder + (*iter));
		
		if(!dir.exists())
		{
			dir.mkdir(file_info.absoluteFilePath() + (*iter));
		}
	}
			
	d_file_log->insertPlainText(QString("\n"));
	this->Add_Log_Entry(QString("Updated Files on %1").arg(QDateTime::currentDateTime().toString("dd MMM yyyy hh:mm:ss AP")));
	
	while(d_TransferFileList.size())
	{
		source_file = QString("%1%2").arg(source_folder).arg(d_TransferFileList[0]);
		destination_file = QString("%1%2").arg(destination_folder).arg(d_TransferFileList[0]);

		if(QFile::exists(destination_file))
		{
			QFile::remove(destination_file);
		}
		
		QFile::copy(source_file,destination_file);
		
		file_info.setFile(source_file);
		date_time = file_info.lastModified();
				
		arguments.clear();
		arguments.push_back(QStringLiteral("-t"));
		arguments.push_back(date_time.toString(d_program_touch_datetime_format));
		arguments.push_back(destination_file);
		
		process.start(d_program_touch,arguments);
		
		process.waitForStarted();
		process.waitForFinished();
		
		this->Add_Log_Entry(date_time.toString("dd MMM yyyy hh:mm:ss AP"),"/" + d_TransferFileList[0]);

		d_TransferFileList.pop_front();
		
		QCoreApplication::processEvents();
	}
		
	return true;
}

void TFolderSync::Add_Transfer_Path(
	const QString 						&path)
{
	QStringList							components;
	int									components_size;
	int									cntr;
	QString								component;
	
	components = path.split('/',QString::SkipEmptyParts);
	
	components_size = components.size();
	
	
	for(cntr = 0;cntr < components_size;cntr++)
	{
		component += QString("/%1").arg(components[cntr]);
		
		if(!d_TransferDirList.contains(component))
		{
			d_TransferDirList.push_back(component);
		}
	}
}

void TFolderSync::Initialize_Log(void)
{
	d_file_log->clear();
}

void TFolderSync::Add_Log_Entry(
	const QString						&title)
{
	d_file_log->moveCursor(QTextCursor::End);
	d_file_log->insertPlainText(QString("%1\n--------------------------------------------\n").arg(title));
}

void TFolderSync::Add_Log_Entry(
	const QString						&date_time,
	const QString						&file)
{
	d_file_log->moveCursor(QTextCursor::End);
	d_file_log->insertPlainText(QString("%1\t%2\n").arg(date_time).arg(file));
}

void TFolderSync::Write_Log(void)
{
	QString								path;
	QFileInfo							file_info;
	QFile								log_file;
	
	path = d_destination_edit->text();
	
	if(path.endsWith('/') || path.endsWith('\\'))
	{
		path.chop(1);
	}
	
	file_info.setFile(path);
	
	if(!file_info.exists())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The target directory does not exist.");
		d_msg_box->setDetailedText(QString("The destination folder '%1' does not exist").arg(file_info.absoluteFilePath()));
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return;
	}
	
	if(!file_info.isDir())
	{
		d_msg_box->setText("Sync Error");
		d_msg_box->setInformativeText("The target directory is not correct");
		d_msg_box->setDetailedText(QString("The destination '%1' must be a directory, not a file").arg(file_info.absoluteFilePath()));
		d_msg_box->setStandardButtons(QMessageBox::Ok);
		d_msg_box->setDefaultButton(QMessageBox::Ok);
		d_msg_box->setIcon(QMessageBox::Critical);
		d_msg_box->exec();
		
		return;
	}
	
	log_file.setFileName(path + QStringLiteral("/controlled document information.txt"));
	
	if(log_file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		log_file.write(d_file_log->toPlainText().toLatin1());
		
		log_file.close();
	}
	
}
